@@ -37,9 +37,12 @@
 #include <stdio.h>
 #include <string.h>
 #include <math.h>
+#include <unistd.h>
+#include <errno.h>
 
 #include "oj.h"
 #include "cache8.h"
+#include "odd.h"
 
 #if !HAS_ENCODING_SUPPORT || defined(RUBINIUS_RUBY)
 #define rb_eEncodingError	rb_eException
@@ -81,10 +84,8 @@ static void	dump_data_comp(VALUE obj, int depth, Out out);
 static void	dump_data_obj(VALUE obj, int depth, Out out);
 static void	dump_obj_comp(VALUE obj, int depth, Out out);
 static void	dump_obj_obj(VALUE obj, int depth, Out out);
-#if HAS_RSTRUCT
 static void	dump_struct_comp(VALUE obj, int depth, Out out);
 static void	dump_struct_obj(VALUE obj, int depth, Out out);
-#endif
 #if HAS_IVAR_HELPERS
 static int	dump_attr_cb(ID key, VALUE value, Out out);
 #endif
@@ -93,6 +94,7 @@ static void	dump_odd(VALUE obj, Odd odd, VALUE clas, int depth, Out out);
 
 static void	grow(Out out, size_t len);
 static size_t	hibit_friendly_size(const uint8_t *str, size_t len);
+static size_t	xss_friendly_size(const uint8_t *str, size_t len);
 static size_t	ascii_friendly_size(const uint8_t *str, size_t len);
 
 static void	dump_leaf(Leaf leaf, int depth, Out out);
@@ -102,6 +104,10 @@ static void	dump_leaf_float(Leaf leaf, Out out);
 static void	dump_leaf_array(Leaf leaf, int depth, Out out);
 static void	dump_leaf_hash(Leaf leaf, int depth, Out out);
 
+// These are used to detect rails re-call of Oj.dump() inside the to_json()
+// method. It is not thread safe.
+static VALUE	last_obj = Qundef;
+static int	oj_rails_hack = -1;
 
 static const char	hex_chars[17] = "0123456789abcdef";
 
@@ -127,6 +133,16 @@ static char	ascii_friendly_chars[256] = "\
 33333333333333333333333333333333\
 33333333333333333333333333333333";
 
+static char	xss_friendly_chars[256] = "\
+66666666222622666666666666666666\
+11211161111111121111111111116161\
+11111111111111111111111111112111\
+11111111111111111111111111111116\
+33333333333333333333333333333333\
+33333333333333333333333333333333\
+33333333333333333333333333333333\
+33333333333333333333333333333333";
+
 inline static size_t
 hibit_friendly_size(const uint8_t *str, size_t len) {
     size_t	size = 0;
@@ -147,9 +163,19 @@ ascii_friendly_size(const uint8_t *str, size_t len) {
     return size - len * (size_t)'0';
 }
 
+inline static size_t
+xss_friendly_size(const uint8_t *str, size_t len) {
+    size_t	size = 0;
+
+    for (; 0 < len; str++, len--) {
+	size += xss_friendly_chars[*str];
+    }
+    return size - len * (size_t)'0';
+}
+
 inline static void
 fill_indent(Out out, int cnt) {
-    if (0 < cnt && 0 < out->indent) {
+    if (0 < out->indent) {
 	cnt *= out->indent;
 	*out->cur++ = '\n';
 	for (; 0 < cnt; cnt--) {
@@ -394,13 +420,13 @@ dump_fixnum(VALUE obj, Out out) {
 
 static void
 dump_bignum(VALUE obj, Out out) {
-    VALUE	rs = rb_big2str(obj, 10);
-    int		cnt = (int)RSTRING_LEN(rs);
+    volatile VALUE	rs = rb_big2str(obj, 10);
+    int			cnt = (int)RSTRING_LEN(rs);
 
     if (out->end - out->cur <= (long)cnt) {
 	grow(out, cnt);
     }
-    memcpy(out->cur, StringValuePtr(rs), cnt);
+    memcpy(out->cur, rb_string_value_ptr((VALUE*)&rs), cnt);
     out->cur += cnt;
     *out->cur = '\0';
 }
@@ -426,10 +452,13 @@ dump_float(VALUE obj, Out out) {
     } else if (-OJ_INFINITY == d) {
 	strcpy(buf, "-Infinity");
 	cnt = 9;
+    } else if (isnan(d)) {
+	strcpy(buf, "NaN");
+	cnt = 3;
     } else if (d == (double)(long long int)d) {
 	cnt = sprintf(buf, "%.1f", d); // used sprintf due to bug in snprintf
     } else {
-	cnt = sprintf(buf, "%0.16g", d); // used sprintf due to bug in snprintf
+	cnt = sprintf(buf, "%0.15g", d); // used sprintf due to bug in snprintf
     }
     if (out->end - out->cur <= (long)cnt) {
 	grow(out, cnt);
@@ -445,10 +474,17 @@ dump_cstr(const char *str, size_t cnt, int is_sym, int escape1, Out out) {
     size_t	size;
     char	*cmap;
 
-    if (Yes == out->opts->ascii_only) {
+    switch (out->opts->escape_mode) {
+    case ASCIIEsc:
 	cmap = ascii_friendly_chars;
 	size = ascii_friendly_size((uint8_t*)str, cnt);
-    } else {
+	break;
+    case XSSEsc:
+	cmap = xss_friendly_chars;
+	size = xss_friendly_size((uint8_t*)str, cnt);
+	break;
+    case JSONEsc:
+    default:
 	cmap = hibit_friendly_chars;
 	size = hibit_friendly_size((uint8_t*)str, cnt);
     }
@@ -518,12 +554,12 @@ dump_cstr(const char *str, size_t cnt, int is_sym, int escape1, Out out) {
 
 static void
 dump_str_comp(VALUE obj, Out out) {
-    dump_cstr(StringValuePtr(obj), RSTRING_LEN(obj), 0, 0, out);
+    dump_cstr(rb_string_value_ptr((VALUE*)&obj), RSTRING_LEN(obj), 0, 0, out);
 }
 
 static void
 dump_str_obj(VALUE obj, Out out) {
-    const char	*s = StringValuePtr(obj);
+    const char	*s = rb_string_value_ptr((VALUE*)&obj);
     size_t	len = RSTRING_LEN(obj);
     char	s1 = s[1];
 
@@ -573,16 +609,14 @@ dump_class_obj(VALUE obj, Out out) {
 
 static void
 dump_array(VALUE a, int depth, Out out) {
-    VALUE	*np;
     size_t	size;
-    int		cnt;
+    int		i, cnt;
     int		d2 = depth + 1;
     long	id = check_circular(a, out);
 
     if (id < 0) {
 	return;
     }
-    np = RARRAY_PTR(a);
     cnt = (int)RARRAY_LEN(a);
     *out->cur++ = '[';
     if (0 < id) {
@@ -612,7 +646,8 @@ dump_array(VALUE a, int depth, Out out) {
 	} else {
 	    size = d2 * out->opts->dump_opts->indent_size + out->opts->dump_opts->array_size + 1;
 	}
-	for (; 0 < cnt; cnt--, np++) {
+	cnt--;
+	for (i = 0; i <= cnt; i++) {
 	    if (out->end - out->cur <= (long)size) {
 		grow(out, size);
 	    }
@@ -631,8 +666,8 @@ dump_array(VALUE a, int depth, Out out) {
 		    }
 		}
 	    }
-	    dump_val(*np, d2, out);
-	    if (1 < cnt) {
+	    dump_val(rb_ary_entry(a, i), d2, out);
+	    if (i < cnt) {
 		*out->cur++ = ',';
 	    }
 	}
@@ -901,6 +936,7 @@ dump_hash(VALUE obj, int depth, int mode, Out out) {
 	    }
 	    if (0 < out->opts->dump_opts->indent_size) {
 		int	i;
+
 		for (i = depth; 0 < i; i--) {
 		    strcpy(out->cur, out->opts->dump_opts->indent);
 		    out->cur += out->opts->dump_opts->indent_size;
@@ -985,9 +1021,9 @@ dump_time(VALUE obj, Out out) {
 
 static void
 dump_ruby_time(VALUE obj, Out out) {
-    VALUE	rstr = rb_funcall(obj, oj_to_s_id, 0);
+    volatile VALUE	rstr = rb_funcall(obj, oj_to_s_id, 0);
 
-    dump_cstr(StringValuePtr(rstr), RSTRING_LEN(rstr), 0, 0, out);
+    dump_cstr(rb_string_value_ptr((VALUE*)&rstr), RSTRING_LEN(rstr), 0, 0, out);
 }
 
 static void
@@ -1083,9 +1119,9 @@ dump_data_strict(VALUE obj, Out out) {
     VALUE	clas = rb_obj_class(obj);
 
     if (oj_bigdecimal_class == clas) {
-	VALUE	rstr = rb_funcall(obj, oj_to_s_id, 0);
+	volatile VALUE	rstr = rb_funcall(obj, oj_to_s_id, 0);
 
-	dump_raw(StringValuePtr(rstr), RSTRING_LEN(rstr), out);
+	dump_raw(rb_string_value_ptr((VALUE*)&rstr), RSTRING_LEN(rstr), out);
     } else {
 	raise_strict(obj);
     }
@@ -1096,9 +1132,9 @@ dump_data_null(VALUE obj, Out out) {
     VALUE	clas = rb_obj_class(obj);
 
     if (oj_bigdecimal_class == clas) {
-	VALUE	rstr = rb_funcall(obj, oj_to_s_id, 0);
+	volatile VALUE	rstr = rb_funcall(obj, oj_to_s_id, 0);
 
-	dump_raw(StringValuePtr(rstr), RSTRING_LEN(rstr), out);
+	dump_raw(rb_string_value_ptr((VALUE*)&rstr), RSTRING_LEN(rstr), out);
     } else {
 	dump_nil(out);
     }
@@ -1106,10 +1142,10 @@ dump_data_null(VALUE obj, Out out) {
 
 static void
 dump_data_comp(VALUE obj, int depth, Out out) {
-    VALUE	o2;
+    volatile VALUE	o2;
 
     if (rb_respond_to(obj, oj_to_hash_id)) {
-	VALUE	h = rb_funcall(obj, oj_to_hash_id, 0);
+	volatile VALUE	h = rb_funcall(obj, oj_to_hash_id, 0);
  
 	if (T_HASH != rb_type(h)) {
 	    rb_raise(rb_eTypeError, "%s.to_hash() did not return a Hash.\n", rb_class2name(rb_obj_class(obj)));
@@ -1117,10 +1153,16 @@ dump_data_comp(VALUE obj, int depth, Out out) {
 	dump_hash(h, depth, out->opts->mode, out);
     } else if (rb_respond_to(obj, oj_as_json_id) && obj != (o2 = rb_funcall(obj, oj_as_json_id, 0))) {
 	dump_val(o2, depth, out);
-    } else if (rb_respond_to(obj, oj_to_json_id)) {
-	VALUE		rs = rb_funcall(obj, oj_to_json_id, 0);
-	const char	*s = StringValuePtr(rs);
-	int		len = (int)RSTRING_LEN(rs);
+    } else if (rb_respond_to(obj, oj_to_json_id) && (!oj_rails_hack || last_obj != obj)) {
+	volatile VALUE	rs;
+	const char	*s;
+	int		len;
+
+	last_obj = obj;
+	rs = rb_funcall(obj, oj_to_json_id, 0);
+	last_obj = Qundef;
+	s = rb_string_value_ptr((VALUE*)&rs);
+	len = (int)RSTRING_LEN(rs);
 
 	if (out->end - out->cur <= len + 1) {
 	    grow(out, len);
@@ -1136,21 +1178,20 @@ dump_data_comp(VALUE obj, int depth, Out out) {
 	    case RubyTime:	dump_ruby_time(obj, out);	break;
 	    case XmlTime:	dump_xml_time(obj, out);	break;
 	    case UnixTime:
-	    default:	dump_time(obj, out);		break;
+	    default:		dump_time(obj, out);		break;
 	    }
 	} else if (oj_bigdecimal_class == clas) {
-	    VALUE	rstr = rb_funcall(obj, oj_to_s_id, 0);
+	    volatile VALUE	rstr = rb_funcall(obj, oj_to_s_id, 0);
 
 	    if (Yes == out->opts->bigdec_as_num) {
-		dump_raw(StringValuePtr(rstr), RSTRING_LEN(rstr), out);
+		dump_raw(rb_string_value_ptr((VALUE*)&rstr), RSTRING_LEN(rstr), out);
 	    } else {
-		dump_cstr(StringValuePtr(rstr), RSTRING_LEN(rstr), 0, 0, out);
+		dump_cstr(rb_string_value_ptr((VALUE*)&rstr), RSTRING_LEN(rstr), 0, 0, out);
 	    }
 	} else {
-	    VALUE	rstr;
+	    volatile VALUE	rstr = rb_funcall(obj, oj_to_s_id, 0);
 
-	    rstr = rb_funcall(obj, oj_to_s_id, 0);
-	    dump_cstr(StringValuePtr(rstr), RSTRING_LEN(rstr), 0, 0, out);
+	    dump_cstr(rb_string_value_ptr((VALUE*)&rstr), RSTRING_LEN(rstr), 0, 0, out);
 	}
     }
 }
@@ -1177,12 +1218,12 @@ dump_data_obj(VALUE obj, int depth, Out out) {
 
 	if (0 == odd) {
 	    if (oj_bigdecimal_class == clas) {
-		VALUE	rstr = rb_funcall(obj, oj_to_s_id, 0);
+		volatile VALUE	rstr = rb_funcall(obj, oj_to_s_id, 0);
 		
 		if (Yes == out->opts->bigdec_as_num) {
-		    dump_raw(StringValuePtr(rstr), RSTRING_LEN(rstr), out);
+		    dump_raw(rb_string_value_ptr((VALUE*)&rstr), RSTRING_LEN(rstr), out);
 		} else {
-		    dump_cstr(StringValuePtr(rstr), RSTRING_LEN(rstr), 0, 0, out);
+		    dump_cstr(rb_string_value_ptr((VALUE*)&rstr), RSTRING_LEN(rstr), 0, 0, out);
 		}
 	    } else {
 		dump_nil(out);
@@ -1196,7 +1237,7 @@ dump_data_obj(VALUE obj, int depth, Out out) {
 static void
 dump_obj_comp(VALUE obj, int depth, Out out) {
     if (rb_respond_to(obj, oj_to_hash_id)) {
-	VALUE	h = rb_funcall(obj, oj_to_hash_id, 0);
+	volatile VALUE	h = rb_funcall(obj, oj_to_hash_id, 0);
  
 	if (T_HASH != rb_type(h)) {
 	    rb_raise(rb_eTypeError, "%s.to_hash() did not return a Hash.\n", rb_class2name(rb_obj_class(obj)));
@@ -1204,10 +1245,16 @@ dump_obj_comp(VALUE obj, int depth, Out out) {
 	dump_hash(h, depth, out->opts->mode, out);
     } else if (rb_respond_to(obj, oj_as_json_id)) {
 	dump_val(rb_funcall(obj, oj_as_json_id, 0), depth, out);
-    } else if (rb_respond_to(obj, oj_to_json_id)) {
-	VALUE		rs = rb_funcall(obj, oj_to_json_id, 0);
-	const char	*s = StringValuePtr(rs);
-	int		len = (int)RSTRING_LEN(rs);
+    } else if (rb_respond_to(obj, oj_to_json_id) && (!oj_rails_hack || last_obj != obj)) {
+	volatile VALUE	rs;
+	const char	*s;
+	int		len;
+
+	last_obj = obj;
+	rs = rb_funcall(obj, oj_to_json_id, 0);
+	last_obj = Qundef;
+	s = rb_string_value_ptr((VALUE*)&rs);
+	len = (int)RSTRING_LEN(rs);
 
 	if (out->end - out->cur <= len + 1) {
 	    grow(out, len);
@@ -1219,17 +1266,17 @@ dump_obj_comp(VALUE obj, int depth, Out out) {
 	VALUE	clas = rb_obj_class(obj);
 
 	if (oj_bigdecimal_class == clas) {
-	    VALUE	rstr = rb_funcall(obj, oj_to_s_id, 0);
+	    volatile VALUE	rstr = rb_funcall(obj, oj_to_s_id, 0);
 
 	    if (Yes == out->opts->bigdec_as_num) {
-		dump_raw(StringValuePtr(rstr), RSTRING_LEN(rstr), out);
+		dump_raw(rb_string_value_ptr((VALUE*)&rstr), RSTRING_LEN(rstr), out);
 	    } else {
-		dump_cstr(StringValuePtr(rstr), RSTRING_LEN(rstr), 0, 0, out);
+		dump_cstr(rb_string_value_ptr((VALUE*)&rstr), RSTRING_LEN(rstr), 0, 0, out);
 	    }
 	} else if (oj_datetime_class == clas || oj_date_class == clas) {
-	    VALUE	rstr = rb_funcall(obj, oj_to_s_id, 0);
+	    volatile VALUE	rstr = rb_funcall(obj, oj_to_s_id, 0);
 
-	    dump_cstr(StringValuePtr(rstr), RSTRING_LEN(rstr), 0, 0, out);
+	    dump_cstr(rb_string_value_ptr((VALUE*)&rstr), RSTRING_LEN(rstr), 0, 0, out);
 	} else {
 	    Odd	odd = oj_get_odd(clas);
 
@@ -1253,9 +1300,9 @@ dump_obj_obj(VALUE obj, int depth, Out out) {
 
 	if (0 == odd) {
 	    if (oj_bigdecimal_class == clas) {
-		VALUE	rstr = rb_funcall(obj, oj_to_s_id, 0);
+		volatile VALUE	rstr = rb_funcall(obj, oj_to_s_id, 0);
 
-		dump_raw(StringValuePtr(rstr), RSTRING_LEN(rstr), out);
+		dump_raw(rb_string_value_ptr((VALUE*)&rstr), RSTRING_LEN(rstr), out);
 	    } else {
 		dump_obj_attrs(obj, clas, id, depth, out);
 	    }
@@ -1303,7 +1350,7 @@ dump_attr_cb(ID key, VALUE value, Out out) {
 
 static void
 dump_obj_attrs(VALUE obj, VALUE clas, slot_t id, int depth, Out out) {
-    size_t	size;
+    size_t	size = 0;
     int		d2 = depth + 1;
 
     if (out->end - out->cur <= 2) {
@@ -1342,11 +1389,10 @@ dump_obj_attrs(VALUE obj, VALUE clas, slot_t id, int depth, Out out) {
     }
     {
 	int	cnt;
-// use encoding as the indicator for Ruby 1.8.7 or 1.9.x
 #if HAS_IVAR_HELPERS
 	cnt = (int)rb_ivar_count(obj);
 #else
-	VALUE		vars = rb_funcall2(obj, oj_instance_variables_id, 0, 0);
+	volatile VALUE	vars = rb_funcall2(obj, oj_instance_variables_id, 0, 0);
 	VALUE		*np = RARRAY_PTR(vars);
 	ID		vid;
 	const char	*attr;
@@ -1395,6 +1441,8 @@ dump_obj_attrs(VALUE obj, VALUE clas, slot_t id, int depth, Out out) {
 #endif
 #if HAS_EXCEPTION_MAGIC
 	if (Qtrue == rb_obj_is_kind_of(obj, rb_eException)) {
+	    volatile VALUE	rv;
+
 	    if (',' != *(out->cur - 1)) {
 		*out->cur++ = ',';
 	    }
@@ -1405,7 +1453,8 @@ dump_obj_attrs(VALUE obj, VALUE clas, slot_t id, int depth, Out out) {
 	    fill_indent(out, d2);
 	    dump_cstr("~mesg", 5, 0, 0, out);
 	    *out->cur++ = ':';
-	    dump_val(rb_funcall2(obj, rb_intern("message"), 0, 0), d2, out);
+	    rv = rb_funcall2(obj, rb_intern("message"), 0, 0);
+	    dump_val(rv, d2, out);
 	    if (out->end - out->cur <= 2) {
 		grow(out, 2);
 	    }
@@ -1417,7 +1466,8 @@ dump_obj_attrs(VALUE obj, VALUE clas, slot_t id, int depth, Out out) {
 	    fill_indent(out, d2);
 	    dump_cstr("~bt", 3, 0, 0, out);
 	    *out->cur++ = ':';
-	    dump_val(rb_funcall2(obj, rb_intern("backtrace"), 0, 0), d2, out);
+	    rv = rb_funcall2(obj, rb_intern("backtrace"), 0, 0);
+	    dump_val(rv, d2, out);
 	    if (out->end - out->cur <= 2) {
 		grow(out, 2);
 	    }
@@ -1425,25 +1475,27 @@ dump_obj_attrs(VALUE obj, VALUE clas, slot_t id, int depth, Out out) {
 #endif
 	out->depth = depth;
     }
+    fill_indent(out, depth);
     *out->cur++ = '}';
     *out->cur = '\0';
 }
 
-#if HAS_RSTRUCT
 static void
 dump_struct_comp(VALUE obj, int depth, Out out) {
     if (rb_respond_to(obj, oj_to_hash_id)) {
-	VALUE	h = rb_funcall(obj, oj_to_hash_id, 0);
+	volatile VALUE	h = rb_funcall(obj, oj_to_hash_id, 0);
  
 	if (T_HASH != rb_type(h)) {
 	    rb_raise(rb_eTypeError, "%s.to_hash() did not return a Hash.\n", rb_class2name(rb_obj_class(obj)));
 	}
 	dump_hash(h, depth, out->opts->mode, out);
     } else if (rb_respond_to(obj, oj_to_json_id)) {
-	VALUE		rs = rb_funcall(obj, oj_to_json_id, 0);
-	const char	*s = StringValuePtr(rs);
-	int		len = (int)RSTRING_LEN(rs);
+	volatile VALUE	rs = rb_funcall(obj, oj_to_json_id, 0);
+	const char	*s;
+	int		len;
 
+	s = rb_string_value_ptr((VALUE*)&rs);
+	len = (int)RSTRING_LEN(rs);
 	if (out->end - out->cur <= len) {
 	    grow(out, len);
 	}
@@ -1458,13 +1510,12 @@ static void
 dump_struct_obj(VALUE obj, int depth, Out out) {
     VALUE	clas = rb_obj_class(obj);
     const char	*class_name = rb_class2name(clas);
-    VALUE	*vp;
     int		i;
     int		d2 = depth + 1;
     int		d3 = d2 + 1;
     size_t	len = strlen(class_name);
     size_t	size = d2 * out->indent + d3 * out->indent + 10 + len;
-	    
+
     if (out->end - out->cur <= (long)size) {
 	grow(out, size);
     }
@@ -1483,28 +1534,48 @@ dump_struct_obj(VALUE obj, int depth, Out out) {
     *out->cur++ = '"';
     *out->cur++ = ',';
     size = d3 * out->indent + 2;
-    for (i = (int)RSTRUCT_LEN(obj), vp = RSTRUCT_PTR(obj); 0 < i; i--, vp++) {
-	if (out->end - out->cur <= (long)size) {
-	    grow(out, size);
+#ifdef RSTRUCT_LEN
+    {
+	VALUE	*vp;
+
+	for (i = (int)RSTRUCT_LEN(obj), vp = RSTRUCT_PTR(obj); 0 < i; i--, vp++) {
+	    if (out->end - out->cur <= (long)size) {
+		grow(out, size);
+	    }
+	    fill_indent(out, d3);
+	    dump_val(*vp, d3, out);
+	    *out->cur++ = ',';
 	}
-	fill_indent(out, d3);
-	dump_val(*vp, d3, out);
-	*out->cur++ = ',';
     }
+#else
+    {
+	// This is a bit risky as a struct in C ruby is not the same as a Struct
+	// class in interpreted Ruby so length() may not be defined.
+	int	slen = FIX2INT(rb_funcall2(obj, oj_length_id, 0, 0));
+
+	for (i = 0; i < slen; i++) {
+	    if (out->end - out->cur <= (long)size) {
+		grow(out, size);
+	    }
+	    fill_indent(out, d3);
+	    dump_val(rb_struct_aref(obj, INT2FIX(i)), d3, out);
+	    *out->cur++ = ',';
+	}
+    }
+#endif
     out->cur--;
     *out->cur++ = ']';
     *out->cur++ = '}';
     *out->cur = '\0';
 }
-#endif
 
 static void
 dump_odd(VALUE obj, Odd odd, VALUE clas, int depth, Out out) {
-    ID		*idp;
-    VALUE	v;
-    const char	*name;
-    size_t	size;
-    int		d2 = depth + 1;
+    ID			*idp;
+    volatile VALUE	v;
+    const char		*name;
+    size_t		size;
+    int			d2 = depth + 1;
 
     if (out->end - out->cur <= 2) {
 	grow(out, 2);
@@ -1612,7 +1683,7 @@ dump_val(VALUE obj, int depth, Out out) {
 	default:		dump_data_obj(obj, depth, out);	break;
 	}
 	break;
-#if HAS_RSTRUCT
+
     case T_STRUCT: // for Range
 	switch (out->opts->mode) {
 	case StrictMode:	raise_strict(obj);		break;
@@ -1622,7 +1693,7 @@ dump_val(VALUE obj, int depth, Out out) {
 	default:		dump_struct_obj(obj, depth, out);	break;
 	}
 	break;
-#endif
+
 #if (defined T_COMPLEX && defined RCOMPLEX)
     case T_COMPLEX:
 #endif
@@ -1657,6 +1728,9 @@ oj_dump_obj_to_json(VALUE obj, Options copts, Out out) {
 	oj_cache8_new(&out->circ_cache);
     }
     out->indent = copts->indent;
+    if (0 > oj_rails_hack) {
+	oj_rails_hack = (rb_const_defined_at(rb_cObject, rb_intern("ActiveSupport")));
+    }
     dump_val(obj, 0, out);
     if (Yes == copts->circular) {
 	oj_cache8_delete(out->circ_cache);
@@ -1669,6 +1743,7 @@ oj_write_obj_to_file(VALUE obj, const char *path, Options copts) {
     struct _Out out;
     size_t	size;
     FILE	*f;
+    int		ok;
 
     out.buf = buf;
     out.end = buf + sizeof(buf) - 10;
@@ -1676,17 +1751,62 @@ oj_write_obj_to_file(VALUE obj, const char *path, Options copts) {
     oj_dump_obj_to_json(obj, copts, &out);
     size = out.cur - out.buf;
     if (0 == (f = fopen(path, "w"))) {
+	if (out.allocated) {
+	    xfree(out.buf);
+	}
 	rb_raise(rb_eIOError, "%s\n", strerror(errno));
     }
-    if (size != fwrite(out.buf, 1, size, f)) {
-	int	err = ferror(f);
-
-	rb_raise(rb_eIOError, "Write failed. [%d:%s]\n", err, strerror(err));
-    }
+    ok = (size == fwrite(out.buf, 1, size, f));
     if (out.allocated) {
 	xfree(out.buf);
     }
     fclose(f);
+    if (!ok) {
+	int	err = ferror(f);
+
+	rb_raise(rb_eIOError, "Write failed. [%d:%s]\n", err, strerror(err));
+    }
+}
+
+void
+oj_write_obj_to_stream(VALUE obj, VALUE stream, Options copts) {
+    char	buf[4096];
+    struct _Out out;
+    ssize_t	size;
+    VALUE	clas = rb_obj_class(stream);
+    VALUE	s;
+
+    out.buf = buf;
+    out.end = buf + sizeof(buf) - 10;
+    out.allocated = 0;
+    oj_dump_obj_to_json(obj, copts, &out);
+    size = out.cur - out.buf;
+    if (oj_stringio_class == clas) {
+	rb_funcall(stream, oj_write_id, 1, rb_str_new(out.buf, size));
+#ifndef JRUBY_RUBY
+#if !IS_WINDOWS
+    } else if (rb_respond_to(stream, oj_fileno_id) && Qnil != (s = rb_funcall(stream, oj_fileno_id, 0))) {
+	int	fd = FIX2INT(s);
+
+	if (size != write(fd, out.buf, size)) {
+	    if (out.allocated) {
+		xfree(out.buf);
+	    }
+	    rb_raise(rb_eIOError, "Write failed. [%d:%s]\n", errno, strerror(errno));
+	}
+#endif
+#endif
+    } else if (rb_respond_to(stream, oj_write_id)) {
+	rb_funcall(stream, oj_write_id, 1, rb_str_new(out.buf, size));
+    } else {
+	if (out.allocated) {
+	    xfree(out.buf);
+	}
+	rb_raise(rb_eArgError, "to_stream() expected an IO Object.");
+    }
+    if (out.allocated) {
+	xfree(out.buf);
+    }
 }
 
 // dump leaf functions
@@ -1708,7 +1828,7 @@ dump_leaf_str(Leaf leaf, Out out) {
 	dump_cstr(leaf->str, strlen(leaf->str), 0, 0, out);
 	break;
     case RUBY_VAL:
-	dump_cstr(StringValuePtr(leaf->value), RSTRING_LEN(leaf->value), 0, 0, out);
+	dump_cstr(rb_string_value_cstr(&leaf->value), RSTRING_LEN(leaf->value), 0, 0, out);
 	break;
     case COL_VAL:
     default:
@@ -1904,3 +2024,130 @@ oj_write_leaf_to_file(Leaf leaf, const char *path, Options copts) {
     }
     fclose(f);
 }
+
+// string writer functions
+
+static void
+key_check(StrWriter sw, const char *key) {
+    DumpType	type = sw->types[sw->depth];
+
+    if (0 == key && (ObjectNew == type || ObjectType == type)) {
+	rb_raise(rb_eStandardError, "Can not push onto an Object without a key.");
+    }
+}
+
+static void
+push_type(StrWriter sw, DumpType type) {
+    if (sw->types_end <= sw->types + sw->depth + 1) {
+	size_t	size = (sw->types_end - sw->types) * 2;
+
+	REALLOC_N(sw->types, char, size);
+	sw->types_end = sw->types + size;
+    }
+    sw->depth++;
+    sw->types[sw->depth] = type;
+}
+
+static void
+maybe_comma(StrWriter sw) {
+    switch (sw->types[sw->depth]) {
+    case ObjectNew:
+	sw->types[sw->depth] = ObjectType;
+	break;
+    case ArrayNew:
+	sw->types[sw->depth] = ArrayType;
+	break;
+    case ObjectType:
+    case ArrayType:
+	// Always have a few characters available in the out.buf.
+	*sw->out.cur++ = ',';
+	break;
+    }
+}
+
+void
+oj_str_writer_push_object(StrWriter sw, const char *key) {
+    size_t	size;
+
+    key_check(sw, key);
+    size = sw->depth * sw->out.indent + 3;
+    if (sw->out.end - sw->out.cur <= size) {
+	grow(&sw->out, size);
+    }
+    maybe_comma(sw);
+    fill_indent(&sw->out, sw->depth);
+    if (0 != key) {
+	dump_cstr(key, strlen(key), 0, 0, &sw->out);
+	*sw->out.cur++ = ':';
+    }
+    *sw->out.cur++ = '{';
+    push_type(sw, ObjectNew);
+}
+
+void
+oj_str_writer_push_array(StrWriter sw, const char *key) {
+    size_t	size;
+
+    key_check(sw, key);
+    size = sw->depth * sw->out.indent + 3;
+    if (sw->out.end - sw->out.cur <= size) {
+	grow(&sw->out, size);
+    }
+    maybe_comma(sw);
+    if (0 != key) {
+	dump_cstr(key, strlen(key), 0, 0, &sw->out);
+	*sw->out.cur++ = ':';
+    }
+    *sw->out.cur++ = '[';
+    push_type(sw, ArrayNew);
+}
+
+void
+oj_str_writer_push_value(StrWriter sw, VALUE val, const char *key) {
+    size_t	size;
+
+    key_check(sw, key);
+    size = sw->depth * sw->out.indent + 3;
+    if (sw->out.end - sw->out.cur <= size) {
+	grow(&sw->out, size);
+    }
+    maybe_comma(sw);
+    if (0 != key) {
+	dump_cstr(key, strlen(key), 0, 0, &sw->out);
+	*sw->out.cur++ = ':';
+    }
+    dump_val(val, sw->depth, &sw->out);
+}
+
+void
+oj_str_writer_pop(StrWriter sw) {
+    size_t	size;
+    DumpType	type = sw->types[sw->depth];
+
+    sw->depth--;
+    if (0 > sw->depth) {
+	rb_raise(rb_eStandardError, "Can not pop with no open array or object.");
+    }
+    size = sw->depth * sw->out.indent + 2;
+    if (sw->out.end - sw->out.cur <= size) {
+	grow(&sw->out, size);
+    }
+    fill_indent(&sw->out, sw->depth);
+    switch (type) {
+    case ObjectNew:
+    case ObjectType:
+	*sw->out.cur++ = '}';
+	break;
+    case ArrayNew:
+    case ArrayType:
+	*sw->out.cur++ = ']';
+	break;
+    }
+}
+
+void
+oj_str_writer_pop_all(StrWriter sw) {
+    while (0 < sw->depth) {
+	oj_str_writer_pop(sw);
+    }
+}
